/*
 *  bignum_div.c
 *  BigMath
 *
 *  Created by Curtis Jones on 2007.08.05.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include "bignum.h"

// TODO: make sure that divs->oofm == 1
//
bignum_t *
bignum_div_aux (bignum_t *quot, bignum_t *rmdr, bignum_t *divd, bignum_t *divs)
{
	u_int64_t r = 0;
	u_int64_t n = divd->oofm;
	u_int64_t j = n - 1;
	u_int64_t b = divd->base;
	
	if (divd->sign == BIGNUM_NEGATIVE) {
		printf("%s().. divd is negative! ", __PRETTY_FUNCTION__); bignum_print(divd);
		return NULL;
	}
	
	if (divs->sign == BIGNUM_NEGATIVE) {
		printf("%s().. divs is negative! ", __PRETTY_FUNCTION__); bignum_print(divs);
		return NULL;
	}
	
	if (quot != NULL)
		__bignum_zero(quot);
	
	do {
		u_int64_t y = r * b + (u_int64_t)divd->numb[j];
		
		if (quot != NULL) {
			__bignum_set(quot, j, y / (u_int64_t)divs->numb[0]);
			
			if (quot->numb[j] != 0)
				quot->oofm = BIGNUM_MAX(quot->oofm, (j+1));
		}
		
		r = y % (u_int64_t)divs->numb[0];
	}
	while (--j >= 0 && j < n);
	
	if (rmdr != NULL) {
		__bignum_set(rmdr, 0, r);
		
		if (rmdr->numb[0] != 0)
			rmdr->oofm = 1;
	}
	
	return quot;
}

// Knuth 4.3.1 - Algorithm D
//
// a / b = c
//
// dividend (a) / divisor (b) = quotient (c)
//
bignum_t *
bignum_div (bignum_t *quot, bignum_t *rmdr, bignum_t *divd, bignum_t *divs)
{
	u_int64_t i = 0;
	u_int64_t m = divd->oofm - divs->oofm;
	u_int64_t n = divs->oofm;
	u_int64_t b = divd->base;
	u_int64_t d = ( b / ((u_int64_t)divs->numb[divs->oofm-1] + 1) );
	bignum_t *u = NULL, *v = NULL;
	u_int32_t divd_sign, divs_sign, quot_sign;
	
	// set divd to positive
	// set divs to positive
	//
	// if divd is negative xor divs is negative
	//   quot is negative
	//
	{
		if (BIGNUM_NEGATIVE == (divd_sign = divd->sign) ^ BIGNUM_NEGATIVE == (divs_sign = divs->sign))
			quot_sign = BIGNUM_NEGATIVE;
		else
			quot_sign = BIGNUM_POSITIVE;
		
		divd->sign = BIGNUM_POSITIVE;
		divs->sign = BIGNUM_POSITIVE;
	}
	
	if (quot != NULL)
		__bignum_zero( quot );
	
	if (rmdr != NULL)
		__bignum_zero( rmdr );
	
	// if the dividend is less than the divisor then the
	// quotient is zero and the remainder is the dividend.
	//
	if (bignum_lt(divd, divs)) {
		if (quot != NULL)
			bignum_set(quot, 0);
		
		if (rmdr != NULL)
			bignum_set_bn(rmdr, divd);
		
		divd->sign = divd_sign;
		divs->sign = divs_sign;
		
		return quot;
	}
	
	// if the dividend is equal to the divisor then the
	// quotient is one and the remainder is zero.
	//
	else if (bignum_eq(divd, divs)) {
		if (quot != NULL)
			bignum_set(quot, 1);
		
		if (rmdr != NULL)
			bignum_set(rmdr, 0);
		
		divd->sign = divd_sign;
		divs->sign = divs_sign;
		
		return quot;
	}
	
	// TODO: It is possible that divs->oofm is greater than
	// one, but that the value is still wholly contained in
	// the first magnitude. we could test the value of divs
	// relative to the maximum value that can be held in
	// a single magnitude....
	//
	else if (divs->oofm == 1) {
		bignum_div_aux(quot, rmdr, divd, divs);
		
		divd->sign = divd_sign;
		divs->sign = divs_sign;
		
		return quot;
	}
	
	if (NULL == (u = bignum_alloc())) {
		printf("%s().. [1] bignum_alloc failed\n", __PRETTY_FUNCTION__);
		return NULL;
	}
	
	if (NULL == (v = bignum_alloc())) {
		printf("%s().. [2] bignum_alloc failed\n", __PRETTY_FUNCTION__);
		bignum_free( u );
		return NULL;
	}
	
	// STEP D1 - Normalize
	//
	// Set u = divisor * d
	// Set v = dividend * d
	//
	// If u is still of the same magnitude, then add
	// an additional zero to the end - the high end.
	// This does not change its value, but the additional
	// space is required.
	//
	{
		bignum_t * tmpd = bignum_alloc();
		
		bignum_set(tmpd, d);
		
		bignum_mul(u, divd, tmpd);
		bignum_mul(v, divs, tmpd);
		
		if (u->oofm == divd->oofm)
			u->oofm++;
		
		bignum_free( tmpd );
	}
	
	// STEP D2 - Initialize j
	//
	// Set j to m which is the difference in magnitude between
	// the dividend and the divisor. We'll be decrementing j
	// each time through this loop.
	//
	{
		u_int64_t j = m;
		
		bignum_t * bn_q    = bignum_alloc();
		bignum_t * bn_tmp1 = bignum_alloc();
		bignum_t * bn_tmp2 = bignum_alloc();
		bignum_t * bn_tmpu = bignum_alloc();
		
		// STEP D3 - Part 1 - Calculate q
		//
		// q is going to be the j'th digit of the quotient. first
		// we estimate q and then possibly alter it slightly if 
		// our estimate is off by one or two.
		//
		do {
			u_int64_t x = (u_int64_t)u->numb[j+n] * b + (u_int64_t)u->numb[j+n-1];
			u_int64_t q = x / (u_int64_t)v->numb[n-1];
			u_int64_t r = x - (q * (u_int64_t)v->numb[n-1]);
			u_int64_t o = 0;
			
			// STEP D3 - Part 2 - Verify q
			//
			// It is possible that the estimate of q is off. This
			// code which I'm not about to explain, alters q as
			// necessary.
			//
			for (i = 0; i < 2; ++i) {
				if (q < b && q * (u_int64_t)v->numb[n-2] <= b * r + u->numb[j+n-2])
					break;
				
				q -= 1;
				r += (u_int64_t)v->numb[n-1];
				
				if (r >= b)
					break;
			}
			
			// STEP D4 - Multiply and Subtract
			//
			// Now that our q correct, multiply "digits" j through
			// j + n of u by v. Then replace "digits" j through 
			// j + n of u with the product. Convoluted subtraction.
			//
			{
				int i = 0;
				
				bignum_set(bn_q, q);
				bignum_mul(bn_tmp1, v, bn_q);
				bignum_set(bn_tmp2, 0);
				bignum_set(bn_tmpu, 0);
				
				for (i = 0; i <= n; ++i)
					__bignum_set(bn_tmpu, bn_tmpu->oofm++, u->numb[j+i]);
				
				if (bignum_lt(bn_tmpu, bn_tmp1)) {
					bignum_set(bn_q, --q);
					bignum_mul(bn_tmp1, v, bn_q);
				}
				
				if (bignum_lt(bn_tmpu, bn_tmp1))
					printf("bignum::bignum_div().. warning: q is still too large.\n");
				
				o = bn_tmpu->oofm;
				bignum_sub(bn_tmp2, bn_tmpu, bn_tmp1);
				bignum_set_bn(bn_tmpu, bn_tmp2);
				
				for (i = 0; i < bn_tmpu->oofm; ++i)
					__bignum_set(u, i+j, bn_tmpu->numb[i]);
				
				for (; i <= n; ++i)
					__bignum_set(u, i+j, 0);
				
				u->oofm -= o - bn_tmpu->oofm;
			}
			
			if (quot != NULL) {
				__bignum_set(quot, j, q);
				
				if (quot->numb[j] != 0)
					quot->oofm = BIGNUM_MAX(quot->oofm, j+1);
			}
		}
		while (--j >= 0 && j <= m);
		
		bignum_free( bn_q    );
		bignum_free( bn_tmp1 );
		bignum_free( bn_tmp2 );
		bignum_free( bn_tmpu );
	}
	
	if (rmdr != NULL) {
		bignum_t *bn_tmpd = bignum_alloc();
		bignum_set(bn_tmpd, d);
		bignum_div_aux(rmdr, NULL, u, bn_tmpd);
		bignum_free( bn_tmpd );
	}
	
	bignum_free( u );
	bignum_free( v );
	
	divd->sign = divd_sign;
	divs->sign = divs_sign;
	
	if (quot != NULL)
		quot->sign = quot_sign;
	
	return quot;
}
