/*
 *  bignum_root.c
 *  BigMath
 *
 *  Created by Curtis Jones on 2008.03.19.
 *  Copyright 2008 __MyCompanyName__. All rights reserved.
 *
 */

#include "bignum.h"

/**
 * Sets root equal to the nth root of num. Returns root.
 *
 * http://en.wikipedia.org/wiki/Nth_root_algorithm
 *
 * 1 / n *
 *   (n - 1) * x_sub_k +
 *   num / (x_sub_k ^ (n - 1))
 *
 * n /
 *   (n - 1) * x_sub_k +
 *   num / (x_sub_k ^ (n - 1))
 *
 * TODO - how do we pick a good first x_sub_k? it seems that using a multi-bit
 *        shift is the way to go for that. we might be fine with a hard-coded
 *        initial x_sub_k for now. adding floating point support might change
 *        that.
 *
 * TODO - special cases:
 *   if num is zero...
 *   if num is negative...
 *   if n is zero...
 *   if n is one... root = num
 *   if n is two...
 *   if n is negative...
 *
 */
bignum_t *
bignum_root (bignum_t *root, bignum_t *n, bignum_t *num)
{
	int first_pass = 0;
	
	bignum_t * result  = NULL;
	bignum_t * x_sub_k = NULL;
	bignum_t * n_min_1 = NULL;
	bignum_t * part_1  = NULL;
	bignum_t * part_2  = NULL;
	bignum_t * part_3  = NULL;
	bignum_t * part_4  = NULL;
	bignum_t * part_5  = NULL;
	bignum_t * part_6  = NULL;
	
	if (NULL == (x_sub_k = bignum_alloc()))
		goto finish;
	
	if (NULL == (n_min_1 = bignum_alloc()))
		goto finish;
	
	if (NULL == (part_1 = bignum_alloc()))
		goto finish;
	
	if (NULL == (part_2 = bignum_alloc()))
		goto finish;
	
	if (NULL == (part_3 = bignum_alloc()))
		goto finish;
	
	if (NULL == (part_4 = bignum_alloc()))
		goto finish;
	
	if (NULL == (part_5 = bignum_alloc()))
		goto finish;
	
	if (NULL == (part_6 = bignum_alloc()))
		goto finish;
	
	bignum_set(x_sub_k, 2);
	
	// n __-__ 1
	bignum_sub(n_min_1, n, bignum_one());
	
	int i = 0;
	
	// we loop forever, and hope that we'll detect that we've found the best 
	// possible answer and break out of the loop. otherwise, there's a short-
	// circuit that'll stop this from looping infinitely.
	//
	while (1)
	{
		// short circuit in case something goes wrong
		if (i++ > 1000)
			break;
		
		// (n - 1) __*__ x_sub_k
		bignum_mul(part_1, n_min_1, x_sub_k);
		
		// x_sub_k __^__ (n - 1)
		bignum_pow(part_2, x_sub_k, n_min_1);
		
		// num __/__ (x_sub_k ^ (n - 1))
		bignum_div(part_3, NULL, num, part_2);
		
		// ((n - 1) * x_sub_k) __+__ (num / (x_sub_k ^ (n - 1)))
		bignum_add(part_4, part_1, part_3);
		
		// [...] __/__ n
		bignum_div(part_5, part_6, part_4, n);
		
		// if the result is zero then something has gone very, very wrong.
		if (bignum_eq(part_5, bignum_zero()))
			break;
		
		// if we're getting a repeating answer, then we have arrived.
		else if (bignum_eq(x_sub_k, part_5))
			break;
		
		// this process should always result in a lower and lower answer, slowly
		// approaching the correct root value. however, if the result is higher 
		// than it was last time, then we have used a value for x_sub_k that is 
		// less than the correct root value.
		//
		// if this is the first time through this process, or this also happened
		// last time, then multiply x_sub_k by one hundred and try again.
		//
		// if this is not the first time through and this did not also happen
		// last time, then we have gone as far as we can go. the current x_sub_k is
		// the answer we're sticking with.
		//
		else if (bignum_lt(x_sub_k, part_5)) {
			if (first_pass || i == 1) {
				bignum_mul(part_6, x_sub_k, bignum_one_hundred());
				bignum_set_bn(x_sub_k, part_6);
				first_pass++;
			}
			else {
				bignum_t * tmp = part_5;
				part_5 = x_sub_k;
				x_sub_k = tmp;
				break;
			}
		}
		else {
			first_pass = 0;
			bignum_set_bn(x_sub_k, part_5);
		}
	}
	
	bignum_set_bn(root, part_5);
	
	result = root;
	
finish:
	if (x_sub_k != NULL)
		bignum_free(x_sub_k);
	
	if (n_min_1 != NULL)
		bignum_free(n_min_1);
	
	if (part_1 != NULL)
		bignum_free(part_1);
	
	if (part_2 != NULL)
		bignum_free(part_2);
	
	if (part_3 != NULL)
		bignum_free(part_3);
	
	if (part_4 != NULL)
		bignum_free(part_4);
	
	if (part_5 != NULL)
		bignum_free(part_5);
	
	if (part_6 != NULL)
		bignum_free(part_6);
	
	return result;
}
